// flip-curses.cpp : Implementation file for flip-curses.h.

#include "flip-curses.h"

#ifdef FLIP_CURSES_SUPPORT

#include "program.h"

#include <curses.h>

#include <ctime>
#include <fstream>
#include <iostream>

static std::ifstream fin;
static std::ofstream fout;

static std::size_t show_lines, show_cols;
static std::vector<std::string> output(1);

static int input_callback_c()
{
    int c = getch();
    if (c == ERR)
        c = -2;
    return c;
}

static int input_callback_f()
{
    return fin.get();
}

static void output_callback_c(int c)
{
    if (c == '\n')
        output.push_back("");
    else if (output.back().size() == (std::size_t)COLS)
        output.push_back(std::string(1, c));
    else
        output.back().push_back((char)c);
    
    if (show_lines + output.size() > (std::size_t)LINES && output.size() > (std::size_t)LINES/2)
        output.erase(output.begin(), output.end() - std::max((std::size_t)LINES - show_lines, (std::size_t)LINES/2));
    if (show_lines + output.size() > (std::size_t)LINES)
        show_lines = LINES - output.size();
}

static void output_callback_f(int c)
{
    fout << (unsigned char)c;
}

template <class SV>
static void updatescreen(const SV& data)
{
    for (std::size_t y = 0; y < show_lines; ++y)
    {
        for (std::size_t x = 0; x < show_cols; ++x)
        {
            if (data[y][x])
                mvaddch(y, x, data[y][x]);
        }
    }
    // TODO: Optimise this to not erase each time.
    move(show_lines, 0);
    clrtobot();
    for (std::size_t y = 0; y < output.size(); ++y)
        mvaddstr(show_lines + y, 0, output[y].c_str());
}

static void periodic_sleep(long delay)
{
    static clock_t goal = 0;
    if (delay == 0)
        goal = clock();
    else
    {
        goal += delay * CLOCKS_PER_SEC / 1000;
        clock_t now;
        while ((now = clock()) - goal < 0);
        goal = now;
    }
}

int flip_curses(const char* pszSourceFile, flip_options options)
{
    program prog;
    std::ifstream source(pszSourceFile);
    prog.load(source);
    source.close();
    
    if (options.infile)
    {
        fin.open(options.infile);
        prog.set_input_callback(input_callback_f);
    }
    else
        prog.set_input_callback(input_callback_c);
    if (options.outfile)
    {
        fout.open(options.outfile);
        prog.set_output_callback(output_callback_f);
    }
    else
        prog.set_output_callback(output_callback_c);
    
    prog.set_unspec_test(options.unspectest);
    
    initscr();
    
    show_lines = std::min(prog.code().size(), (std::size_t)LINES-output.size());
    show_cols = std::min(prog.code()[0].size(), (std::size_t)COLS);
    
    if (!options.outfile)
    {
        noecho();
        nodelay(stdscr, TRUE);
    }
    
    erase();
    updatescreen(prog.code());
    refresh();
    
    periodic_sleep(0);
    
    int result;
    while ((result = prog.step()) == 0)
    {
        updatescreen(prog.updates());
        refresh();
        periodic_sleep(options.speed);
    }
    
    endwin();
    
    if (options.infile)
        fin.close();
    if (options.outfile)
        fout.close();
    
    switch (result)
    {
        case 1:
            result = 0;
            break;
        case 2:
            std::cerr << "All balls are trapped or have left the program." << std::endl;
            break;
        case 3:
            std::cerr << "Unspecified behaviour detected." << std::endl;
            break;
        default:
        {
            program::ball_t error_ball = prog.get_error_ball();
            std::cerr << "Invalid interaction between ball (" << error_ball.dir << ") and object '"
                      << prog.code()[error_ball.y][error_ball.x] << "' ("
                      << (int)(unsigned char)prog.code()[error_ball.y][error_ball.x]
                      << ") at line " << error_ball.y+1 << ", column " << error_ball.x+1 << "."
                      << std::endl;
            break;
        }
    }
    
    return result;
}

#else

int flip_curses(const char* /*pszSourceFile*/, flip_options /*options*/)
{
    return 1;
}

#endif
