// program.c : Implementation file for program.h.

#include "program.h"

#include <algorithm>
#include <cctype>
#include <cstring>
#include <functional>
#include <sstream>

// Program elements.
// Most of theses macros are unused, but they make a nice reference.
#define MIRRORS    "-|"
#define FLIPPERS   "/\\"
#define SLUCES     "><^v"
#define NUMBERGENS "0123456789"
#define TARPITS    "+*"
#define UNARY      "~',."
#define IO         "pPrR"
#define MISC       "Q#X"
#define LEVELS     "ZN"
#define OBJECTS    MIRRORS FLIPPERS SLUCES NUMBERGENS TARPITS UNARY IO MISC LEVELS
#define MODIFIERS  "@+-0~%"

#define MOD_NW 1
#define MOD_NE 2
#define MOD_SW 4
#define MOD_SE 8
#define MOD_EVAL(values, pos1, pos2) (((mod & pos1) != 0) ^ ((mod & pos2) != 0))

program::program()
    : tabstop(8), input(NULL), output(NULL), unspectest(false), error_iball(-1)
{
}

void program::load(std::istream& in)
{
    source.clear();
    std::string line;
    size_t maxlen = 1;
    while (std::getline(in, line))
    {
        std::size_t i = 0;
        while ((i = line.find_first_of('\t', i)) != std::string::npos)
            line.replace(i, 1, tabstop - i%tabstop, ' ');
        line.resize(line.find_last_not_of(' ') + 1);
        if (line.size() > maxlen)
            maxlen = line.size();
        source.push_back(line);
    }
    if (source.empty())
        source.push_back("");
    
    for (std::vector<std::string>::iterator i = source.begin(), last = source.end(); i != last; ++i)
    {
        i->resize(maxlen, ' ');
        screendata.push_back(std::vector<char>(maxlen, '\0'));
    }
    
    ball_t initialball = {-1, 0, 0, '>', 0};
    balls.clear();
    balls.push_back(initialball);
    
    extradata_index_map.resize(source.size());
    c_flippers = c_tarpits = 0;
    for (std::size_t i=0; i<source.size(); ++i)
    {
        extradata_index_map[i].resize(maxlen);
        for (std::size_t j = source[i].find_first_of(FLIPPERS); j != std::string::npos; j = source[i].find_first_of(FLIPPERS, j+1))
            extradata_index_map[i][j] = c_flippers++;
        for (std::size_t j = source[i].find_first_of(TARPITS); j != std::string::npos; j = source[i].find_first_of(TARPITS, j+1))
            extradata_index_map[i][j] = c_tarpits++;
    }
    
    flipper_states.clear();
    tarpit_states.clear();
}

struct used_flipper_data
{
    std::size_t x, y, level;
    bool flipped;
    bool operator ==(const used_flipper_data& rhs) const { return x == rhs.x && y == rhs.y && level == rhs.level; }
};

int program::step()
{
// Priority order, i.e. a lower value trumps higher ones.
#define RF_ERROR   1
#define RF_UNSPEC  2
#define RF_NOBALLS 4
#define RF_QUIT    8
    int result_flags = 0;
    
    // For unspecified behaviour detection.
    bool used_input = false, used_output = false;
    std::vector<used_flipper_data> used_flippers;
    
    for (std::vector<std::vector<char> >::iterator i = screendata.begin(), last = screendata.end(); i != last; ++i)
        std::fill(i->begin(), i->end(), '\0');
    
    for (std::vector<ball_t>::iterator iball = balls.begin(); iball != balls.end(); )
    {
        if ((int)iball->x >= 0 && screendata[iball->y][iball->x] == '\0')
        {
            // Reveal the character under the ball.
            char c = source[iball->y][iball->x];
            if (std::strchr(FLIPPERS, c))
            {
                // Show the state of the flipper on level 0.
                c = flipper_state(iball->x, iball->y, 0);
            }
            else if (std::strchr(TARPITS, c))
            {
                // Show the lowest trapped ball.
                for (std::size_t level = 0; level < tarpit_states.size(); ++level)
                {
                    tarstate_t& ts = tarpit_state(iball->x, iball->y, level);
                    if (ts.has_ball)
                    {
                        c = ball_char(ts.value);
                        break;
                    }
                }
            }
            screendata[iball->y][iball->x] = c;
        }
        switch (iball->dir)
        {
            case '>':
                if (iball->x == source[0].size()-1)
                {
                    iball = balls.erase(iball);
                    continue;
                }
                else
                    iball->x++;
                break;
            case '<':
                if (iball->x == 0)
                {
                    iball = balls.erase(iball);
                    continue;
                }
                else
                    iball->x--;
                break;
            case '^':
                if (iball->y == 0)
                {
                    iball = balls.erase(iball);
                    continue;
                }
                else
                    iball->y--;
                break;
            case 'v':
                if (iball->y == source.size()-1)
                {
                    iball = balls.erase(iball);
                    continue;
                }
                else
                    iball->y++;
                break;
        }
        char sourcechar = source[iball->y][iball->x];
        switch (sourcechar)
        {
            case ' ':
                break;
            case '-':
                switch (iball->dir)
                {
                    case '>':
                    case '<':
                        if (!(result_flags & RF_ERROR))
                            error_iball = iball - balls.begin();
                        result_flags |= RF_ERROR;
                        break;
                    case '^':
                        iball->dir = 'v';
                        break;
                    case 'v':
                        iball->dir = '^';
                        break;
                }
                break;
            case '|':
                switch (iball->dir)
                {
                    case '>':
                        iball->dir = '<';
                        break;
                    case '<':
                        iball->dir = '>';
                        break;
                    case '^':
                    case 'v':
                        if (!(result_flags & RF_ERROR))
                            error_iball = iball - balls.begin();
                        result_flags |= RF_ERROR;
                        break;
                }
                break;
            case '/':
            case '\\':
            {
                int mod = modifier_values(iball->x, iball->y, iball->value);
                bool flipped;
                if (flipper_state(iball->x, iball->y, iball->level) == '/')
                {
                    switch (iball->dir)
                    {
                        case '>': iball->dir = '^'; break;
                        case '<': iball->dir = 'v'; break;
                        case '^': iball->dir = '>'; break;
                        case 'v': iball->dir = '<'; break;
                    }
                    if (flipped = MOD_EVAL(mod, MOD_NE, MOD_SW))
                        flipper_state(iball->x, iball->y, iball->level) = '\\';
                }
                else
                {
                    switch (iball->dir)
                    {
                        case '>': iball->dir = 'v'; break;
                        case '<': iball->dir = '^'; break;
                        case '^': iball->dir = '<'; break;
                        case 'v': iball->dir = '>'; break;
                    }
                    if (flipped = MOD_EVAL(mod, MOD_NW, MOD_SE))
                        flipper_state(iball->x, iball->y, iball->level) = '/';
                }
                if (unspectest)
                {
                    used_flipper_data test = {iball->x, iball->y, iball->level, flipped};
                    std::vector<used_flipper_data>::iterator it = std::find(used_flippers.begin(), used_flippers.end(), test);
                    if (it == used_flippers.end())
                        used_flippers.push_back(test);
                    else
                        if (flipped || it->flipped)
                            result_flags |= RF_UNSPEC;
                }
                break;
            }
            case '>':
                if (iball->dir == '<')
                {
                    int mod = modifier_values(iball->x, iball->y, iball->value);
                    bool mod_n = MOD_EVAL(mod, MOD_NW, MOD_NE);
                    bool mod_s = MOD_EVAL(mod, MOD_SW, MOD_SE);
                    if (mod_n && !mod_s) iball->dir = '^';
                    else if (mod_s && !mod_n) iball->dir = 'v';
                    else iball->dir = '>';
                }
                else
                    iball->dir = '>';
                break;
            case '<':
                if (iball->dir == '>')
                {
                    int mod = modifier_values(iball->x, iball->y, iball->value);
                    bool mod_n = MOD_EVAL(mod, MOD_NW, MOD_NE);
                    bool mod_s = MOD_EVAL(mod, MOD_SW, MOD_SE);
                    if (mod_n && !mod_s) iball->dir = '^';
                    else if (mod_s && !mod_n) iball->dir = 'v';
                    else iball->dir = '<';
                }
                else
                    iball->dir = '<';
                break;
            case '^':
                if (iball->dir == 'v')
                {
                    int mod = modifier_values(iball->x, iball->y, iball->value);
                    bool mod_w = MOD_EVAL(mod, MOD_NW, MOD_SW);
                    bool mod_e = MOD_EVAL(mod, MOD_NE, MOD_SE);
                    if (mod_w && !mod_e) iball->dir = '<';
                    else if (mod_e && !mod_w) iball->dir = '>';
                    else iball->dir = '^';
                }
                else
                    iball->dir = '^';
                break;
            case 'v':
                if (iball->dir == '^')
                {
                    int mod = modifier_values(iball->x, iball->y, iball->value);
                    bool mod_w = MOD_EVAL(mod, MOD_NW, MOD_SW);
                    bool mod_e = MOD_EVAL(mod, MOD_NE, MOD_SE);
                    if (mod_w && !mod_e) iball->dir = '<';
                    else if (mod_e && !mod_w) iball->dir = '>';
                    else iball->dir = 'v';
                }
                else
                    iball->dir = 'v';
                break;
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            {
                std::size_t i = iball - balls.begin();
                ball_t newball = {iball->x, iball->y, iball->level, iball->dir, sourcechar - '0'};
                balls.push_back(newball);
                iball = balls.begin() + i;
                switch (iball->dir)
                {
                    case '>': iball->dir = '<'; break;
                    case '<': iball->dir = '>'; break;
                    case '^': iball->dir = 'v'; break;
                    case 'v': iball->dir = '^'; break;
                }
                break;
            }
            case '+':
            case '*':
            {
                tarstate_t& ts = tarpit_state(iball->x, iball->y, iball->level);
                if (ts.has_ball)
                {
                    if (sourcechar == '+')
                        iball->value += ts.value;
                    else
                        iball->value *= ts.value;
                    ts.has_ball = false;
                }
                else
                {
                    ts.has_ball = true;
                    ts.value = iball->value;
                    screendata[iball->y][iball->x] = ball_char(ts.value);
                    iball = balls.erase(iball);
                    continue;
                }
                break;
            }
            case '~':
                iball->value = -iball->value;
                break;
            case '\'':
                iball->value++;
                break;
            case ',':
                iball->value--;
                break;
            case '.':
                iball->value = 0;
                break;
            case 'p':
                if (output)
                {
                    std::ostringstream strm;
                    strm << iball->value << ' ';
                    std::string str = strm.str();
                    for (std::string::iterator i = str.begin(); i != str.end(); ++i)
                        output(*i);
                }
                iball = balls.erase(iball);
                if (unspectest)
                    if (used_output)
                        result_flags |= RF_UNSPEC;
                    else
                        used_output = true;
                continue;
            case 'P':
                if (output)
                    output((unsigned char)iball->value);
                iball = balls.erase(iball);
                if (unspectest)
                    if (used_output)
                        result_flags |= RF_UNSPEC;
                    else
                        used_output = true;
                continue;
            case 'r':
            case 'R':
                if (unspectest)
                    if (used_input)
                        result_flags |= RF_UNSPEC;
                    else
                        used_input = true;
                if (iball->value == 0)
                {
                    if (!(sourcechar == 'r' ? input_int(iball->value) : input_char(iball->value)))
                    {
                        // Erase ball.
                        iball = balls.erase(iball);
                        continue;
                    }
                }
                else if (iball->value > 0)
                {
                    if (sourcechar == 'r' ? input_int(iball->value) : input_char(iball->value))
                    {
                        // Unsuspend ball.
                        switch (iball->dir)
                        {
                            case 'e': iball->dir = '>'; break;
                            case 'w': iball->dir = '<'; break;
                            case 'n': iball->dir = '^'; break;
                            case 's': iball->dir = 'v'; break;
                        }
                    }
                    else
                    {
                        // Suspend ball.
                        switch (iball->dir)
                        {
                            case '>': iball->dir = 'e'; break;
                            case '<': iball->dir = 'w'; break;
                            case '^': iball->dir = 'n'; break;
                            case 'v': iball->dir = 's'; break;
                        }
                    }
                }
                else // if (iball->value < 0)
                {
                    // Suspend whole program until input is ready.
                    while (!(sourcechar == 'r' ? input_int(iball->value) : input_char(iball->value)))
                        ;
                }
                break;
            case 'Q':
                result_flags |= RF_QUIT;
                if (unspectest)
                    if (used_output)
                        result_flags |= RF_UNSPEC;
                    else
                        used_output = true;
                break;
            case '#':
                if (iball->value <= 0)
                {
                    iball = balls.erase(iball);
                    continue;
                }
                break;
            case 'X':
            {
                int mod = modifier_values(iball->x, iball->y, iball->value);
                ball_t ball = *iball;
                switch (iball->dir)
                {
                    case '>':
                    case '<':
                        if (!MOD_EVAL(mod, MOD_NW, MOD_NE))
                        {
                            ball.dir = '^';
                            iball = balls.insert(iball, ball);
                            ++iball;
                        }
                        if (!MOD_EVAL(mod, MOD_SW, MOD_SE))
                        {
                            ball.dir = 'v';
                            iball = balls.insert(iball, ball);
                            ++iball;
                        }
                        break;
                    case '^':
                    case 'v':
                        if (!MOD_EVAL(mod, MOD_NW, MOD_SW))
                        {
                            ball.dir = '<';
                            iball = balls.insert(iball, ball);
                            ++iball;
                        }
                        if (!MOD_EVAL(mod, MOD_NE, MOD_SE))
                        {
                            ball.dir = '>';
                            iball = balls.insert(iball, ball);
                            ++iball;
                        }
                        break;
                }
                screendata[iball->y][iball->x] = ball_char(iball->value);
                iball = balls.erase(iball);
                continue;
            }
            case 'Z':
                iball->level++;
                break;
            case 'N':
                if (iball->level > 0)
                    iball->level--;
                else
                    switch (iball->dir)
                    {
                        case '>': iball->dir = '<'; break;
                        case '<': iball->dir = '>'; break;
                        case '^': iball->dir = 'v'; break;
                        case 'v': iball->dir = '^'; break;
                    }
                break;
            default:
                if (!(result_flags & RF_ERROR))
                    error_iball = iball - balls.begin();
                result_flags |= RF_ERROR;
                break;
        }
        
        screendata[iball->y][iball->x] = ball_char(iball->value);
        ++iball;
    }
    
    if (balls.empty())
        result_flags |= RF_NOBALLS;
    
    if (result_flags & RF_ERROR)
        return -1;
    else if (result_flags & RF_UNSPEC)
        return 3;
    else if (result_flags & RF_NOBALLS)
        return 2;
    else if (result_flags & RF_QUIT)
        return 1;
    else
        return 0;
}

static bool is_digit(char c)
{
    return std::isdigit(c) != 0;
}

static bool is_not_digit(char c)
{
    return std::isdigit(c) == 0;
}

bool program::input_int_from_queue(ball_t::value_type& dst)
{
    if (inputqueue.empty())
        return false;
    std::deque<char>::iterator queuefirst = inputqueue.begin(), queuelast = inputqueue.end(), i, j;
    
    if ((i = std::find_if(queuefirst, queuelast, is_digit)) != queuelast &&
        (j = std::find_if(i, queuelast, is_not_digit)) != queuelast)
    {
        if (i != queuefirst && i[-1] == '-')
            --i;
        
        dst = std::atoi(std::string(i, j).c_str());
        inputqueue.erase(queuefirst, j);
        return true;
    }
    else
        return false;
}

bool program::input_int(ball_t::value_type& dst)
{
    while (!input_int_from_queue(dst))
    {
        if (!input)
            return false;
        int c = input();
        if (c < 0)
            return false;
        inputqueue.push_back((char)c);
    }
    return true;
}

bool program::input_char(ball_t::value_type& dst)
{
    if (!inputqueue.empty())
    {
        dst = inputqueue.front();
        inputqueue.pop_front();
        return true;
    }
    else if (input)
    {
        int c = input();
        if (c == -2)
            return false;
        dst = c;
        return true;
    }
    else
        return false;
}

char program::ball_char(ball_t::value_type value)
{
    int c = value % 10;
    if (c < 0)
        c = -c;
    return (char)('0' + c);
}

char& program::flipper_state(std::size_t x, std::size_t y, std::size_t level)
{
    while (flipper_states.size() <= level)
        flipper_states.push_back(std::vector<char>(c_flippers, '\0'));
    
    if (flipper_states[level][extradata_index_map[y][x]] == '\0')
        flipper_states[level][extradata_index_map[y][x]] = source[y][x];
    return flipper_states[level][extradata_index_map[y][x]];
}

program::tarstate_t& program::tarpit_state(std::size_t x, std::size_t y, std::size_t level)
{
    const tarstate_t empty = {false, 0};
    while (tarpit_states.size() <= level)
        tarpit_states.push_back(std::vector<tarstate_t>(c_tarpits, empty));
    
    return tarpit_states[level][extradata_index_map[y][x]];
}

int program::modifier_values(std::size_t x, std::size_t y, ball_t::value_type ball_value)
{
    const bool values[] = {
        true,
        ball_value > 0,
        ball_value < 0,
        ball_value == 0,
        (ball_value & 1) != 0,
        (rand() & 16) != 0
    };
    const char modifiers[] = MODIFIERS;
    
    int result = 0;
    if (y > 0)
    {
        if (x > 0)
        {
            const char* p = std::strchr(modifiers, source[y-1][x-1]);
            if (p != NULL && values[p-modifiers])
                result |= MOD_NW;
        }
        if (x < source[0].size() - 1)
        {
            const char* p = std::strchr(modifiers, source[y-1][x+1]);
            if (p != NULL && values[p-modifiers])
                result |= MOD_NE;
        }
    }
    if (y < source.size() - 1)
    {
        if (x > 0)
        {
            const char* p = std::strchr(modifiers, source[y+1][x-1]);
            if (p != NULL && values[p-modifiers])
                result |= MOD_SW;
        }
        if (x < source[0].size() - 1)
        {
            const char* p = std::strchr(modifiers, source[y+1][x+1]);
            if (p != NULL && values[p-modifiers])
                result |= MOD_SE;
        }
    }
    
    return result;
}
