// program.h : Interface for the program class.

#ifndef PROGRAM_H
#define PROGRAM_H

#if __cplusplus >= 201103L
// C++11 compilers should have <cstdint>.
# include <cstdint>
# define BALL_VALUE_TYPE std::int32_t
#elif defined(_MSC_VER)
// MSC has fixed width types __intX.
# define BALL_VALUE_TYPE __int32
#else
// Otherwise we hope for a C99 <stdint.h>.
# include <stdint.h>
# define BALL_VALUE_TYPE int32_t
#endif

#include <deque>
#include <istream>
#include <string>
#include <vector>

class program
{
public:
    program();
    
        // Types.
    // Return a value in the range [0-255] normally, -1 for EOF, -2 for no input available.
    typedef int (*input_t)();
    // Output a character in the range [0-255].
    typedef void (*output_t)(int c);
    
    struct ball_t
    {
        typedef BALL_VALUE_TYPE value_type;
        std::size_t x, y, level;
        char        dir;
        value_type  value;
    };
    
        // Setup.
    void set_tabstops(std::size_t tabstop_length) { tabstop = tabstop_length; }
    void load(std::istream& in);
    void set_input_callback(input_t input_callback) { input = input_callback; }
    void set_output_callback(output_t output_callback) { output = output_callback; }
    void set_unspec_test(bool unspec_test) { unspectest = unspec_test; }
    
        // Running.
    // Return values:
    //  0 Normal execution.
    //  1 Ball hit 'Q' (normal exit).
    //  2 No balls remaining.
    //  3 Unspecified behaviour detected (if set_unspec_test(true) was called).
    // -1 Error.
    int step();
    
        // Access.
    const std::vector<std::string>& code() const { return source; }
    const std::vector<std::vector<char> >& updates() const { return screendata; }
    
        // Error reporting.
    ball_t get_error_ball() const { return balls.at(error_iball); }
    
private:
        // Types.
    struct tarstate_t
    {
        bool               has_ball;
        ball_t::value_type value;
    };
    
        // Program source.
    std::vector<std::string>               source;
    std::vector<std::vector<std::size_t> > extradata_index_map;
    std::size_t                            c_flippers;
    std::size_t                            c_tarpits;
    
        // Program state.
    std::vector<ball_t>                   balls;
    std::vector<std::vector<char> >       flipper_states;
    std::vector<std::vector<tarstate_t> > tarpit_states;
    
        // Other stuff.
    std::size_t                     tabstop;
    input_t                         input;
    output_t                        output;
    bool                            unspectest;
    std::vector<std::vector<char> > screendata;
    std::deque<char>                inputqueue;
    std::size_t                     error_iball;
    
        // Methods.
    bool        input_int_from_queue(ball_t::value_type& dst);
    bool        input_int(ball_t::value_type& dst);
    bool        input_char(ball_t::value_type& dst);
    void        set_error(ball_t const& ball);
    static char ball_char(ball_t::value_type value);
    char&       flipper_state(std::size_t x, std::size_t y, std::size_t level);
    tarstate_t& tarpit_state(std::size_t x, std::size_t y, std::size_t level);
    int         modifier_values(std::size_t x, std::size_t y, ball_t::value_type ball_value);
};

#endif
